/* main.c
 * Command line interface.
 *
 * This fixed point version of shine is based on Gabriel Bouvigne's original
 * source, version 0.1.2
 * It was converted for use on Acorn computers running RISC OS.
 * 09/02/01 P.Everett
 */

/*int errno;*/

#include <string.h>
#include <signal.h>
#include "types.h"

config_t config;
int raw;
int mono_from_stereo;
int cutoff;
static int processor_defined;
static int strongarm;

volatile int pleaseDieNow=0;

int (*mul)(int a, int b);
int (*muls)(int a, int b);
int (*mulr)(int a, int b);
int (*mulsr)(int a, int b);
void (*L3_window_filter_subband)(unsigned long **buffer, long s[SBLIMIT], int k);
void (*L3_mdct_sub)(long sb_sample[2][3][18][SBLIMIT],
                 long mdct_freq[2][2][samp_per_frame2]);

extern int mul_a(int a, int b);
extern int muls_a(int a, int b);
extern int mulr_a(int a, int b);
extern int mulsr_a(int a, int b);
extern int mul_sa(int a, int b);
extern int muls_sa(int a, int b);
extern int mulr_sa(int a, int b);
extern int mulsr_sa(int a, int b);
extern void L3_window_filter_subband_a(unsigned long **buffer, long s[SBLIMIT], int k);
extern void L3_mdct_sub_a(long sb_sample[2][3][18][SBLIMIT],
                 long mdct_freq[2][2][samp_per_frame2]);
extern void L3_window_filter_subband_sa(unsigned long **buffer, long s[SBLIMIT], int k);
extern void L3_mdct_sub_sa(long sb_sample[2][3][18][SBLIMIT],
                 long mdct_freq[2][2][samp_per_frame2]);


void handleSignal(int signal)
{
	pleaseDieNow = 1;
}

/*
 * error:
 * ------
 */
void error(char *s)
{
  printf("[ERROR] %s\n",s);
  exit(1);
}

/*
 * print_usage:
 * ------------
 */
static void print_usage()
{
  printf("\nUSAGE   :  Shine [options] <infile> <outfile>\n");
  printf("options : -h            this help message\n");
  printf("          -b <bitrate>  set the bitrate [32-320], default 128kbit\n");
  printf("          -c            set copyright flag, default off\n");
  printf("          -o            reset original flag, default on\n");
  printf("          -r            raw cd data file instead of wave\n");
  printf("          -m            mono from stereo, raw mono with -r\n");
  printf("     The following are for test purposes:\n");
  printf("          -ps           force stongarm code\n");
  printf("          -pa           force normal arm code\n");
  printf("\n");
  exit(0);
}

/*
 * set_functions
 * -------------
 * Initialises the function pointers for strongarm or normal arm
 * function versions.
 */
static void set_functions(int strongarm)
{
    mul   = mul_sa;
    muls  = muls_sa;
    mulr  = mulr_sa;
    mulsr = mulsr_sa;
    L3_window_filter_subband = L3_window_filter_subband_sa;
    L3_mdct_sub = L3_mdct_sub_sa;
}

/*
 * set_defaults:
 * -------------
 */
static void set_defaults()
{
  config.mpeg.type = 1;
  config.mpeg.layr = LAYER_3;
  config.mpeg.mode = MODE_DUAL_CHANNEL;
  config.mpeg.bitr = 128;
  config.mpeg.psyc = 0;
  config.mpeg.emph = 0;
  config.mpeg.crc  = 0;
  config.mpeg.ext  = 0;
  config.mpeg.mode_ext  = 0;
  config.mpeg.copyright = 0;
  config.mpeg.original  = 1;
  cutoff = 418; /* 16KHz @ 44.1Ksps */
}

/*
 * parse_command:
 * --------------
 */
static int parse_command(int argc, char** argv)
{
  int i = 0;

  if(argc<3) return false;

  while(argv[++i][0]=='-')
    switch(argv[i][1])
    {
      case 'b':
        config.mpeg.bitr = atoi(argv[++i]);
        break;

      case 'c':
        config.mpeg.copyright = 1;
        break;

      case 'o':
        config.mpeg.original = 0;
        break;

      case 'r':
        raw = true;
        break;

      case 'm':
        mono_from_stereo = true;
        config.mpeg.mode = MODE_MONO;
        break;

      case 'p':
        switch(argv[i][2])
        {
          case 's':
            strongarm = true;
          case 'a':
            processor_defined = true;
            break;

          default:
            return false;
        }
        break;

      case 'h':
      default :
        return false;
    }

  if((argc-i)!=2) return false;
  config.infile  = argv[i++];
  config.outfile = argv[i];
  return true;
}

/*
 * find_samplerate_index:
 * ----------------------
 */
static int find_samplerate_index(long freq)
{
  static long mpeg1[3] = {44100, 48000, 32000};
  int i;

  for(i=0;i<3;i++)
    if(freq==mpeg1[i]) return i;

  error("Invalid samplerate");
  return 0;
}

/*
 * find_bitrate_index:
 * -------------------
 */
static int find_bitrate_index(int bitr)
{
  static long mpeg1[15] = {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320};
  int i;

  for(i=0;i<15;i++)
    if(bitr==mpeg1[i]) return i;

  error("Invalid bitrate");
  return 0;
}

void set_cutoff(void)
{
  static int cutoff_tab[3][2][15] =
  {
    { /* 44.1k */
      {100,104,131,157,183,209,261,313,365,418,418,418,418,418,418}, /* stereo */
      {183,209,261,313,365,418,418,418,418,418,418,418,418,418,418}  /* mono */
    },
    { /* 48k */
      {100,104,131,157,183,209,261,313,384,384,384,384,384,384,384}, /* stereo */
      {183,209,261,313,365,384,384,384,384,384,384,384,384,384,384}  /* mono */
    },
    { /* 32k */
      {100,104,131,157,183,209,261,313,365,418,522,576,576,576,576}, /* stereo */
      {183,209,261,313,365,418,522,576,576,576,576,576,576,576,576}  /* mono */
    }
  };

  cutoff = cutoff_tab[config.mpeg.samplerate_index]
                     [config.mpeg.mode == MODE_MONO]
                     [config.mpeg.bitrate_index];
}

/*
 * check_config:
 * -------------
 */
static void check_config()
{
  static char *mode_names[4]    = { "stereo", "j-stereo", "dual-ch", "mono" };
  static char *layer_names[4]   = { "", "III", "II", "I" };
  static char *version_names[2] = { "MPEG-II (LSF)", "MPEG-I" };
  static char *psy_names[3]     = { "none", "MUSICAM", "Shine" };
  static char *demp_names[4]    = { "none", "50/15us", "", "CITT" };

  config.mpeg.samplerate_index = find_samplerate_index(config.wave.samplerate);
  config.mpeg.bitrate_index    = find_bitrate_index(config.mpeg.bitr);

  printf("%s layer %s, %s  Psychoacoustic Model: %s\n",
           version_names[config.mpeg.type],
           layer_names[config.mpeg.layr],
           mode_names[config.mpeg.mode],
           psy_names[config.mpeg.psyc]);
  printf("Bitrate=%d kbps  ",config.mpeg.bitr );
  printf("De-emphasis: %s   %s %s\n",
          demp_names[config.mpeg.emph],
          ((config.mpeg.original)?"Original":""),
          ((config.mpeg.copyright)?"(C)":""));
}

/*
 * main:
 * -----
 */
int main(int argc, char **argv)
{
  time_t end_time;
  int
    filetype,
    ok;

  time(&config.start_time);
  set_defaults();
  ok = parse_command(argc,argv);

  printf("ARM Shine v1.06 (32 bit beta) 17/12/02");

  if(!ok)
    print_usage();

  set_functions(strongarm);

  filetype = WAVE;
  if (strncmp("/dev", config.infile, 4) == 0) {
	filetype = OSS;
  }

  if(raw)
    wave_open(DATA); /* prints wave (input) configuration */
  else
    switch(filetype)
    {
      case CD_DATA:
      case WAVE:
      case DATA:
      case OSS:
        wave_open(filetype);
        break;

      default: wave_open(WAVE);
    }

  check_config(); /* prints mpeg (output) configuration */

  set_cutoff();

  printf("Encoding \"%s\" to \"%s\"\n", config.infile, config.outfile);

  if (filetype==OSS) {
  	signal(SIGHUP, handleSignal);
  	signal(SIGINT, handleSignal);
  	signal(SIGQUIT, handleSignal);
  	signal(SIGTERM, handleSignal);
  }
  L3_compress();

  wave_close();

  time(&end_time);
  end_time -= config.start_time;
  printf(" Finished in %2d:%2d:%2d\n",
            end_time/3600,(end_time/60)%60,end_time%60);
  exit(0);
}

